// Fill out your copyright notice in the Description page of Project Settings.

#pragma once

#include "CoreMinimal.h"
#include "Components/ActorComponent.h"
#include "Game/SOCGameState.h"

#include "BuoyancyComponent.generated.h"

USTRUCT(BlueprintType)
struct FVertexData
{
	GENERATED_BODY()

	FVector GlobalPosition;
	
	// depth of the vertex above the surface of the water
	// negative value means the vertex is submerged
	float Depth;

	int32 Index;

	FVertexData() {}

	FVertexData(FVector Position)
	{
		this->GlobalPosition = Position;
		Depth = 0.0f;
	}

	FVertexData(FVector Position, float Depth)
	{
		this->GlobalPosition = Position;
		this->Depth = Depth;
	}
	
	void CalculateDepth(float SurfaceHeight)
	{
		Depth = GlobalPosition.Z - SurfaceHeight;
	}

	bool operator< (const FVertexData& rhs) const
	{
		return this->Depth < rhs.Depth;
	}
};

USTRUCT(BlueprintType)
struct FTriangleData
{
	GENERATED_BODY()
	
	FVertexData V1; // lowest
	FVertexData V2; // middle
	FVertexData V3; // highest

	FVector Center;
	
	float DistanceToSurface;

	FVector Normal;

	float Area;
	
	FTriangleData() {}
	
	FTriangleData(const FVertexData V1, const FVertexData V2, const FVertexData V3, float OceanHeight)
	{
		this->V1 = V1;
		this->V2 = V2;
		this->V3 = V3;

		Center = (V1.GlobalPosition + V2.GlobalPosition + V3.GlobalPosition) / 3.0f;

		DistanceToSurface = FMath::Abs(OceanHeight - Center.Z);
		
		Normal = FVector::CrossProduct(V2.GlobalPosition - V1.GlobalPosition, V3.GlobalPosition - V1.GlobalPosition);
		Normal.Normalize();

		// Heron's formula
		float V12 = FVector::Distance(V1.GlobalPosition, V2.GlobalPosition);
		float V23 = FVector::Distance(V2.GlobalPosition, V3.GlobalPosition);
		float V13 = FVector::Distance(V1.GlobalPosition, V3.GlobalPosition);

		// Half Triangle Perimeter
		float S = (V12 + V23 + V13) / 2.0f;
		
		Area = FMath::Sqrt(S * (S - V12) * (S - V23) * (S - V13));
	}
	
	int32 GetNumSubmergedVertices() const
	{
		int32 result = 0;
		if(V3.Depth < 0.0f)
		{
			result++;
		}
		if(V2.Depth < 0.0f)
		{
			result++;
		}
		if(V1.Depth < 0.0f)
		{
			result++;
		}
		return result;
	}
};

UCLASS( ClassGroup=(Custom), meta=(BlueprintSpawnableComponent) )
class SEAOFCROOKS_API UBuoyancyComponent : public UActorComponent
{
	GENERATED_BODY()

public:	
	// Sets default values for this component's properties
	UBuoyancyComponent();
	
protected:
	UPROPERTY(BlueprintReadOnly)
	UStaticMesh* VolumeMesh;

	UPROPERTY(BlueprintReadOnly)
	TArray<FTriangleData> SubmergedTriangles;

	AOcean* Ocean;
	
	// Called when the game starts
	virtual void BeginPlay() override;

public:	
	// Called every frame
	virtual void TickComponent(float DeltaTime, ELevelTick TickType, FActorComponentTickFunction* ThisTickFunction) override;

private:
	UPrimitiveComponent* VolumeAsPrimComp;
	
	int32 GetNumTriangles();
	
	TArray<FVertexData> GetVertexPositions();

	TArray<int32> GetTriangleIndices();

	void ProcessTriangles(TArray<FTriangleData>& Triangles);
};
